unit uUndoManager;

interface

uses
  System.Classes, System.Generics.Collections;

type
  TUndoStack = TObjectStack<TStream>;
  TUndoStreamEvent = procedure(Stream: TStream) of object;

  TUndoManager = class
  private
    FDisableCount: Integer;
    FOnLoadFromUndoStream: TUndoStreamEvent;
    FOnSaveToUndoStream: TUndoStreamEvent;
    FRedoStack: TUndoStack;
    FUndoLevel: Integer;
    FUndoStack: TUndoStack;
    function GetRedoAvailable: Boolean;
    function GetUndoAvailable: Boolean;
  protected
    procedure DoLoadFromUndoStream(Stream: TStream);
    procedure DoSaveToUndoStream(Stream: TStream);
    procedure InternalLoadFromStream(Stream: TStream);
    procedure InternalSaveToStream(Stream: TStream);
    procedure PopData(Source: TUndoStack);
    procedure PushData(Target: TUndoStack);
  public
    constructor Create;
    destructor Destroy; override;
    procedure BeginUndo;
    procedure CancelUndo;
    procedure Clear;
    procedure DisableUndo;
    procedure EnableUndo;
    procedure EndUndo;
    procedure Redo;
    procedure Undo;
    property RedoAvailable: Boolean read GetRedoAvailable;
    property UndoAvailable: Boolean read GetUndoAvailable;
    property OnLoadFromUndoStream: TUndoStreamEvent read FOnLoadFromUndoStream write FOnLoadFromUndoStream;
    property OnSaveToUndoStream: TUndoStreamEvent read FOnSaveToUndoStream write FOnSaveToUndoStream;
  end;

implementation

uses
  System.SysUtils;

constructor TUndoManager.Create;
begin
  inherited Create;
  FUndoStack := TUndoStack.Create();
  FRedoStack := TUndoStack.Create();
end;

destructor TUndoManager.Destroy;
begin
  FreeAndNil(FRedoStack);
  FreeAndNil(FUndoStack);
  inherited Destroy;
end;

procedure TUndoManager.BeginUndo;
begin
  if (FUndoLevel = 0) and (FDisableCount = 0) then begin
    PushData(FUndoStack);
  end;
  Inc(FUndoLevel);
end;

procedure TUndoManager.CancelUndo;
begin
  Dec(FUndoLevel);
  if (FUndoLevel = 0) and (FDisableCount = 0) then begin
    FUndoStack.Pop;
  end;
end;

procedure TUndoManager.Clear;
begin
  FDisableCount := 0;
  FUndoLevel := 0;
  FUndoStack.Clear;
  FRedoStack.Clear;
end;

procedure TUndoManager.DisableUndo;
begin
  Inc(FDisableCount);
end;

procedure TUndoManager.DoLoadFromUndoStream(Stream: TStream);
begin
  if Assigned(FOnLoadFromUndoStream) then FOnLoadFromUndoStream(Stream);
end;

procedure TUndoManager.DoSaveToUndoStream(Stream: TStream);
begin
  if Assigned(FOnSaveToUndoStream) then FOnSaveToUndoStream(Stream);
end;

procedure TUndoManager.EnableUndo;
begin
  Dec(FDisableCount);
end;

procedure TUndoManager.EndUndo;
begin
  Dec(FUndoLevel);
  if (FUndoLevel = 0) and (FDisableCount = 0) then begin
    FRedoStack.Clear;
  end;
end;

procedure TUndoManager.Redo;
begin
  PushData(FUndoStack);
  PopData(FRedoStack);
end;

function TUndoManager.GetRedoAvailable: Boolean;
begin
  Result := (FUndoLevel = 0) and (FRedoStack.Count > 0);
end;

procedure TUndoManager.Undo;
begin
  PushData(FRedoStack);
  PopData(FUndoStack);
end;

function TUndoManager.GetUndoAvailable: Boolean;
begin
  Result := (FUndoLevel = 0) and (FUndoStack.Count > 0);
end;

procedure TUndoManager.InternalLoadFromStream(Stream: TStream);
begin
  DoLoadFromUndoStream(Stream);
end;

procedure TUndoManager.InternalSaveToStream(Stream: TStream);
begin
  DoSaveToUndoStream(Stream);
end;

procedure TUndoManager.PopData(Source: TUndoStack);
begin
  DisableUndo;
  try
    InternalLoadFromStream(Source.Peek);
  finally
    EnableUndo;
  end;
  Source.Pop;
end;

procedure TUndoManager.PushData(Target: TUndoStack);
var
  stream: TMemoryStream;
begin
  stream := TMemoryStream.Create;
  InternalSaveToStream(stream);
  stream.Position := 0;
  Target.Push(stream);
end;

end.
